import logging
from argparse import ArgumentParser, Namespace
from os import environ
from typing import Any, List

from rich.console import Console

YARALYZE = 'yaralyze'
YARALYZER = f"{YARALYZE}r".upper()
PYTEST_FLAG = 'INVOKED_BY_PYTEST'

KILOBYTE = 1024
MEGABYTE = 1024 * KILOBYTE


def config_var_name(env_var: str) -> str:
    """
    Get the name of env_var and strip off 'YARALYZER_', e.g.:
        SURROUNDING_BYTES_ENV_VAR = 'YARALYZER_SURROUNDING_BYTES'
        config_var_name(SURROUNDING_BYTES_ENV_VAR) => 'SURROUNDING_BYTES'
    """
    env_var = env_var.removeprefix("YARALYZER_")
    return f'{env_var=}'.partition('=')[0]


def is_env_var_set_and_not_false(var_name):
    """Returns True if var_name is not empty and set to anything other than 'false' (capitalization agnostic)"""
    if var_name in environ:
        var_value = environ[var_name]
        return var_value is not None and len(var_value) > 0 and var_value.lower() != 'false'
    else:
        return False


def is_invoked_by_pytest():
    """Return true if pytest is running"""
    return is_env_var_set_and_not_false(PYTEST_FLAG)


class YaralyzerConfig:
    # Passed through to yara.set_config()
    DEFAULT_MAX_MATCH_LENGTH = 100 * KILOBYTE
    DEFAULT_YARA_STACK_SIZE = 2 * 65536

    # Skip decoding binary matches under/over these lengths
    DEFAULT_MIN_DECODE_LENGTH = 1
    DEFAULT_MAX_DECODE_LENGTH = 256

    # chardet.detect() related
    DEFAULT_MIN_CHARDET_TABLE_CONFIDENCE = 2
    DEFAULT_MIN_CHARDET_BYTES = 9

    # Number of bytes to show before/after byte previews and decodes. Configured by command line or env var
    DEFAULT_SURROUNDING_BYTES = 64

    LOG_DIR_ENV_VAR = 'YARALYZER_LOG_DIR'
    LOG_DIR = environ.get(LOG_DIR_ENV_VAR)
    LOG_LEVEL_ENV_VAR = f"{YARALYZER}_LOG_LEVEL"
    LOG_LEVEL = logging.getLevelName(environ.get(LOG_LEVEL_ENV_VAR, 'WARN'))

    if LOG_DIR and not is_invoked_by_pytest():
        Console(color_system='256').print(f"Writing logs to '{LOG_DIR}' instead of stderr/stdout...", style='dim')

    HIGHLIGHT_STYLE = 'orange1'

    ONLY_CLI_ARGS = [
        'debug',
        'help',
        'hex_patterns',
        'interact',
        'patterns_label',
        'regex_patterns',
        'regex_modifier',
        'version'
    ]

    @classmethod
    def set_argument_parser(cls, parser: ArgumentParser) -> None:
        cls._argument_parser: ArgumentParser = parser
        cls._argparse_keys: List[str] = sorted([action.dest for action in parser._actions])

    @classmethod
    def set_args(cls, args: Namespace) -> None:
        cls.args = args

        for option in cls._argparse_keys:
            if option.startswith('export') or option in cls.ONLY_CLI_ARGS:
                continue

            arg_value = vars(args)[option]
            env_var = f"{YARALYZER}_{option.upper()}"
            env_value = environ.get(env_var)
            default_value = cls.get_default_arg(option)
            #print(f"option: {option}, arg_value: {arg_value}, env_var: {env_var}, env_value: {env_value}, default: {default_value}")

            # TODO: as is you can't override env vars with CLI args
            if isinstance(arg_value, bool):
                setattr(args, option, arg_value or is_env_var_set_and_not_false(env_var))
            elif isinstance(arg_value, (int, float)):
                # Check against defaults to avoid overriding env var configured options
                if arg_value == default_value and env_value is not None:
                    setattr(args, option, int(env_value) or arg_value)  # TODO: float args not handled
            else:
                setattr(args, option, arg_value or env_value)

    @classmethod
    def set_default_args(cls):
        cls.set_args(cls._argument_parser.parse_args(['dummy']))

    @classmethod
    def get_default_arg(cls, arg: str) -> Any:
        default_var = f"DEFAULT_{arg.upper()}"
        return vars(cls).get(default_var)



"""
Variables and methods for working with Rich text output.
"""
from shutil import get_terminal_size
from sys import exit
from typing import List

from rich import box
from rich.console import Console
from rich.errors import MarkupError
from rich.panel import Panel
from rich.style import Style
from rich.text import Text
from rich.theme import Theme

from yaralyzer.config import is_env_var_set_and_not_false, is_invoked_by_pytest

DEFAULT_CONSOLE_WIDTH = 160

# Colors
ALERT_STYLE = 'error'  # Regex Capture used when extracting quoted chunks of bytes
BYTES = 'color(100) dim'
BYTES_NO_DIM = 'color(100)'
BYTES_BRIGHTEST = 'color(220)'
BYTES_BRIGHTER = 'orange1'
BYTES_HIGHLIGHT = 'color(136)'
DANGER_HEADER = 'color(88) on white'  # Red
DARK_GREY = 'color(236)'
GREY = 'color(241)'
GREY_ADDRESS = 'color(238)'
PEACH = 'color(215)'

# Theme used by main console
YARALYZER_THEME_DICT = {
    # colors
    'dark_orange': 'color(58)',
    'grey': GREY,
    'grey.dark': DARK_GREY,
    'grey.dark_italic': f"{DARK_GREY} italic",
    'grey.darker_italic': 'color(8) dim italic',
    'grey.darkest': 'color(235) dim',
    'grey.light': 'color(248)',
    'off_white': 'color(245)',
    'zero_bytes': 'color(20)',
    # data types
    'encoding': 'color(158) underline bold',
    'encoding.header': 'color(158) bold',
    'encoding.language': 'dark_green italic',
    'number': 'cyan',
    'regex': 'color(218) dim',
    'no_attempt': "color(60) dim italic",
    # design elements
    'decode.section_header': 'color(100) reverse',
    'decode.subheading': PEACH,
    'decode.subheading_2': 'color(215) dim italic',
    'decode.table_header': 'color(101) bold',
    'headline': 'bold white underline',
    # bytes
    'ascii': 'color(58)',
    'ascii_unprintable': 'color(131)',
    'bytes': BYTES,
    'bytes.title_dim': 'orange1 dim',
    'bytes.title': BYTES_BRIGHTER,
    'bytes.decoded': BYTES_BRIGHTEST,
    # yara
    'matched_rule': 'on bright_black bold',
    'yara.key': DARK_GREY,
    'yara.match_var': 'color(156) italic',
    'yara.string': 'white',
    'yara.date': 'color(216)',
    'yara.url': 'color(220)',
    'yara.int': 'color(45)',
    'yara.hex': 'color(98)',
    'yara.scanned': Style(color='yellow', underline=True, bold=True),
    'yara.rules':  Style(color='color(135)', underline=True, bold=True),
    # error log events
    'error': 'bright_red',
}

YARALYZER_THEME = Theme(YARALYZER_THEME_DICT)


def console_width_possibilities():
    # Subtract 2 from terminal cols just as a precaution in case things get weird
    return [get_terminal_size().columns - 2, DEFAULT_CONSOLE_WIDTH]


def console_width() -> int:
    """Current width set in console obj"""
    return console._width or 40


# Maximize output width if YARALYZER_MAXIMIZE_WIDTH is set (also can changed with --maximize-width option)
if is_invoked_by_pytest():
    CONSOLE_WIDTH = DEFAULT_CONSOLE_WIDTH
elif is_env_var_set_and_not_false('YARALYZER_MAXIMIZE_WIDTH'):
    CONSOLE_WIDTH = max(console_width_possibilities())
else:
    CONSOLE_WIDTH = min(console_width_possibilities())

# Many bytes take 4 chars to print (e.g. '\xcc') so this is the max bytes we can safely print in a line
CONSOLE_PRINT_BYTE_WIDTH = int(CONSOLE_WIDTH / 4.0)
console = Console(theme=YARALYZER_THEME, color_system='256', highlight=False, width=CONSOLE_WIDTH)


def console_print_with_fallback(_string, style=None) -> None:
    """Fallback to regular print() if there's a Markup issue"""
    try:
        console.print(_string, style=style)
    except MarkupError:
        console.print(f"Hit a bracket issue with rich.console printing, defaulting to plain print", style='warn')
        print(_string.plain if isinstance(_string, Text) else _string)


def theme_colors_with_prefix(prefix: str) -> List[Text]:
    return [Text(k, v) for k, v in YARALYZER_THEME.styles.items() if k.startswith(prefix)]


def print_fatal_error_and_exit(error_message: str) -> None:
    console.line(1)
    print_header_panel(error_message, style='bold red reverse')
    console.line(1)
    exit()


def print_header_panel(headline: str, style: str, expand: bool = True, padding: tuple = (0,2)) -> None:
    console.print(Panel(headline, box=box.DOUBLE_EDGE, style=style, expand=expand, padding=padding))



"""
Class to handle attempting to decode a chunk of bytes into strings with various possible encodings.
Leverages the chardet library to both guide what encodings are attempted as well as to rank decodings
in the results.
"""

from collections import defaultdict
from copy import deepcopy
from operator import attrgetter
from typing import List, Optional

from rich.align import Align
from rich.console import Console, ConsoleOptions, NewLine, RenderResult
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

#from yaralyzer.bytes_match import BytesMatch
from yaralyzer.config import YaralyzerConfig
from yaralyzer.decoding.decoding_attempt import DecodingAttempt
from yaralyzer.encoding_detection.character_encodings import ENCODING, ENCODINGS_TO_ATTEMPT, encoding_offsets
from yaralyzer.encoding_detection.encoding_assessment import EncodingAssessment
from yaralyzer.encoding_detection.encoding_detector import EncodingDetector
from yaralyzer.helpers.dict_helper import get_dict_key_by_value
from yaralyzer.helpers.list_helper import flatten
from yaralyzer.helpers.rich_text_helper import CENTER, DECODING_ERRORS_MSG, NO_DECODING_ERRORS_MSG
from yaralyzer.output.decoding_attempts_table import (DecodingTableRow, assessment_only_row,
     decoding_table_row, new_decoding_attempts_table)
from yaralyzer.util.logging import log

# A 2-tuple that can be indexed by booleans of messages used in the table to show true vs. false
WAS_DECODABLE_YES_NO = [NO_DECODING_ERRORS_MSG, DECODING_ERRORS_MSG]

# Multiply chardet scores by 100 (again) to make sorting the table easy
SCORE_SCALER = 100.0


class BytesDecoder:
    def __init__(self, bytes_match: 'BytesMatch', label: Optional[str] = None) -> None:
        self.bytes_match = bytes_match
        self.bytes = bytes_match.surrounding_bytes
        self.label = label or bytes_match.label

        # Empty table/metrics/etc
        self.was_match_decodable = _build_encodings_metric_dict()
        self.was_match_force_decoded = _build_encodings_metric_dict()
        self.was_match_undecodable = _build_encodings_metric_dict()
        self.decoded_strings = {}  # dict[encoding: decoded string]
        self.undecoded_rows = []
        self.decodings = []

        # Note we send both the match and surrounding bytes used when detecting the encoding
        self.encoding_detector = EncodingDetector(self.bytes)

    def __rich_console__(self, _console: Console, options: ConsoleOptions) -> RenderResult:
        """Rich object generator (see Rich console docs)"""
        yield NewLine(2)
        yield Align(self._decode_attempt_subheading(), CENTER)

        if not YaralyzerConfig.args.suppress_chardet:
            yield NewLine()
            yield Align(self.encoding_detector, CENTER)
            yield NewLine()

        # In standalone mode we always print the hex/raw bytes
        if self.bytes_match.is_decodable():
            yield self._build_decodings_table()
        elif YaralyzerConfig.args.standalone_mode:
            # TODO: yield self.bytes_match.suppression_notice() (i guess to show some notice that things are suppressed?)
            yield self._build_decodings_table(True)

        yield NewLine()
        yield Align(self.bytes_match.bytes_hashes_table(), CENTER, style='dim')

    def _build_decodings_table(self, suppress_decodes: bool = False) -> Table:
        """First rows are the raw / hex views of the bytes, next rows are the attempted decodings"""
        self.table = new_decoding_attempts_table(self.bytes_match)

        # Add the encoding rows to the table if not suppressed
        if not (YaralyzerConfig.args.suppress_decoding_attempts or suppress_decodes):
            self.decodings = [DecodingAttempt(self.bytes_match, encoding) for encoding in ENCODINGS_TO_ATTEMPT]
            # Attempt decodings we don't usually attempt if chardet is insistent enough
            forced_decodes = self._undecoded_assessments(self.encoding_detector.force_decode_assessments)
            self.decodings += [DecodingAttempt(self.bytes_match, a.encoding) for a in forced_decodes]

            # If we still haven't decoded chardet's top choice, decode it
            if len(self._forced_displays()) > 0 and not self._was_decoded(self._forced_displays()[0].encoding):
                chardet_top_encoding = self._forced_displays()[0].encoding
                log.info(f"Decoding {chardet_top_encoding} because it's chardet top choice...")
                self.decodings.append(DecodingAttempt(self.bytes_match, chardet_top_encoding))

            # Build the table rows from the decoding attempts
            rows = [self._row_from_decoding_attempt(decoding) for decoding in self.decodings]
            rows += [assessment_only_row(a, a.confidence * SCORE_SCALER) for a in self._forced_displays()]
            self._track_decode_stats()

            for row in sorted(rows, key=attrgetter('sort_score', 'encoding_label_plain'), reverse=True):
                self.table.add_row(*row[0:4])

        return self.table

    # TODO: rename this to something that makes more sense, maybe assessments_over_display_threshold()?
    def _forced_displays(self) -> List[EncodingAssessment]:
        """Returns assessments over the display threshold that are not yet decoded."""
        return self._undecoded_assessments(self.encoding_detector.force_display_assessments)

    def _undecoded_assessments(self, assessments: List[EncodingAssessment]) -> List[EncodingAssessment]:
        """Filter out the already decoded assessments from a set of assessments"""
        return [a for a in assessments if not self._was_decoded(a.encoding)]

    def _was_decoded(self, encoding: str) -> bool:
        """Check whether a given encoding is in the table already"""
        return any(row.encoding == encoding for row in self.decodings)

    def _decode_attempt_subheading(self) -> Panel:
        """Generate a rich.Panel for displaying decode attempts"""
        headline = Text(f"Found ", style='decode.subheading') + self.bytes_match.__rich__()
        return Panel(headline, style='decode.subheading', expand=False)

    def _track_decode_stats(self) -> None:
        """Track stats about successful vs. forced vs. failed decode attempts"""
        for decoding in self.decodings:
            if decoding.failed_to_decode:
                self.was_match_undecodable[decoding.encoding] += 1
                continue

            self.was_match_decodable[decoding.encoding] += 1

            if decoding.was_force_decoded:
                self.was_match_force_decoded[decoding.encoding] += 1

    def _row_from_decoding_attempt(self, decoding: DecodingAttempt) -> DecodingTableRow:
        """Create a DecodingAttemptTable row from a DecodingAttempt."""
        assessment = self.encoding_detector.get_encoding_assessment(decoding.encoding)

        # If the decoding can have a start offset add an appropriate extension to the encoding label
        if decoding.start_offset_label:
            if assessment.language:
                log.warning(f"{decoding.encoding} has offset {decoding.start_offset} and language '{assessment.language}'")
            else:
                assessment = deepcopy(assessment)
                assessment.set_encoding_label(decoding.start_offset_label)

        plain_decoded_string = decoding.decoded_string.plain
        sort_score = assessment.confidence * SCORE_SCALER

        # If the decoding result is a duplicate of a previous decoding, replace the decoded text
        # with "same output as X" where X is the previous encoding that gave the same result.
        if plain_decoded_string in self.decoded_strings.values():
            encoding_with_same_output = get_dict_key_by_value(self.decoded_strings, plain_decoded_string)
            display_text = Text('same output as ', style='color(66) dim italic')
            display_text.append(encoding_with_same_output, style=ENCODING).append('...', style='white')
        else:
            self.decoded_strings[decoding.encoding_label] = plain_decoded_string
            display_text = decoding.decoded_string

        # Set failures negative, shave off a little for forced decodes
        if decoding.failed_to_decode:
            sort_score = (sort_score * -1) - 100
        elif decoding.was_force_decoded:
            sort_score -= 10

        was_forced = WAS_DECODABLE_YES_NO[int(decoding.was_force_decoded)]
        return decoding_table_row(assessment, was_forced, display_text, sort_score)


def _build_encodings_metric_dict():
    """One key for each key in ENCODINGS_TO_ATTEMPT, values are all 0"""
    metrics_dict = defaultdict(lambda: 0)

    for encoding in ENCODINGS_TO_ATTEMPT.keys():
        metrics_dict[encoding] = 0

    return metrics_dict



"""
Help with lists.
"""

def flatten(a):
    """From https://www.geeksforgeeks.org/python/python-flatten-list-to-individual-elements/"""
    return_value = []

    for x in a:
        if isinstance(x, list):
            return_value.extend(flatten(x))  # Recursively flatten nested lists
        else:
            return_value.append(x)  # Append individual elements

    return return_value



import code
from os import environ, getcwd, path

from dotenv import load_dotenv
from rich.text import Text

# load_dotenv() should be called as soon as possible (before parsing local classes) but not for pytest
if not environ.get('INVOKED_BY_PYTEST', False):
    for dotenv_file in [path.join(dir, '.yaralyzer') for dir in [getcwd(), path.expanduser('~')]]:
        if path.exists(dotenv_file):
            load_dotenv(dotenv_path=dotenv_file)
            break

from yaralyzer.config import YaralyzerConfig
from yaralyzer.output.file_export import export_json, invoke_rich_export
from yaralyzer.output.rich_console import console
from yaralyzer.util.argument_parser import get_export_basepath, parse_arguments
from yaralyzer.util.logging import log
from yaralyzer.yara.yara_rule_builder import HEX, REGEX
from yaralyzer.yaralyzer import Yaralyzer

PDFALYZER_MSG = "\nIf you are analyzing a PDF you may be interested in The Pdfalyzer, birthplace of The Yaralyzer:"
PDFALYZER_MSG_TXT = Text(PDFALYZER_MSG, style='bright_white bold')
PDFALYZER_MSG_TXT.append('\n -> ', style='bright_white')
PDFALYZER_MSG_TXT.append('https://github.com/michelcrypt4d4mus/pdfalyzer\n', style='bright_cyan underline')


def yaralyze():
    args = parse_arguments()
    output_basepath = None

    if args.yara_rules_files:
        yaralyzer = Yaralyzer.for_rules_files(args.yara_rules_files, args.file_to_scan_path)
    elif args.yara_rules_dirs:
        yaralyzer = Yaralyzer.for_rules_dirs(args.yara_rules_dirs, args.file_to_scan_path)
    elif args.regex_patterns or args.hex_patterns:
        yaralyzer = Yaralyzer.for_patterns(
            args.regex_patterns or args.hex_patterns,
            HEX if args.hex_patterns else REGEX,
            args.file_to_scan_path,
            pattern_label=args.patterns_label,
            regex_modifier=args.regex_modifier)
    else:
        raise RuntimeError("No pattern or YARA file to scan against.")

    if args.output_dir:
        output_basepath = get_export_basepath(args, yaralyzer)
        console.print(f"Will render yaralyzer data to '{output_basepath}'...", style='yellow')
        console.record = True

    yaralyzer.yaralyze()

    if args.export_txt:
        invoke_rich_export(console.save_text, output_basepath)

    if args.export_html:
        invoke_rich_export(console.save_html, output_basepath)

    if args.export_svg:
        invoke_rich_export(console.save_svg, output_basepath)

    if args.export_json:
        export_json(yaralyzer, output_basepath)

    if args.file_to_scan_path.endswith('.pdf'):
        console.print(PDFALYZER_MSG_TXT)

    # Drop into interactive shell if requested
    if args.interact:
        code.interact(local=locals())



"""
Helper methods to work with bytes.
"""
import re
from io import StringIO
from sys import byteorder

from rich.console import Console
from rich.markup import escape
from rich.text import Text

from yaralyzer.bytes_match import BytesMatch
from yaralyzer.config import YaralyzerConfig
from yaralyzer.encoding_detection.character_encodings import NEWLINE_BYTE, encoding_width
from yaralyzer.helpers.rich_text_helper import newline_join
from yaralyzer.output.rich_console import (BYTES, BYTES_BRIGHTER, BYTES_BRIGHTEST,
     BYTES_HIGHLIGHT, GREY, console, console_width)
from yaralyzer.util.logging import log

HEX_CHARS_PER_GROUP = 8
SUBTABLE_MAX_WIDTH = console_width() - 35 - 5  # 35 for first 3 cols, 5 for in between hex and ascii
HEX_UNIT_LENGTH = (HEX_CHARS_PER_GROUP * 3) + HEX_CHARS_PER_GROUP + 4  # 4 for padding between groups
HEX_GROUPS_PER_LINE = divmod(SUBTABLE_MAX_WIDTH, HEX_UNIT_LENGTH)[0]
HEX_CHARS_PER_LINE = HEX_CHARS_PER_GROUP * HEX_GROUPS_PER_LINE


def get_bytes_before_and_after_match(_bytes: bytes, match: re.Match, num_before=None, num_after=None) -> bytes:
    """
    Get all bytes from num_before the start of the sequence up until num_after the end of the sequence
    num_before and num_after will both default to the env var/CLI options having to do with surrounding
    bytes. If only num_before is provided then num_after will use it as a default.
    """
    return get_bytes_surrounding_range(_bytes, match.start(), match.end(), num_before, num_after)


def get_bytes_surrounding_range(_bytes: bytes, start_idx: int, end_idx: int, num_before=None, num_after=None) -> bytes:
    num_after = num_after or num_before or YaralyzerConfig.args.surrounding_bytes
    num_before = num_before or YaralyzerConfig.args.surrounding_bytes
    start_idx = max(start_idx - num_before, 0)
    end_idx = min(end_idx + num_after, len(_bytes))
    return _bytes[start_idx:end_idx]


def clean_byte_string(bytes_array: bytes) -> str:
    """Gives you a string representation of bytes w/no cruft e.g. '\x80\nx44' instead of "b'\x80\nx44'"."""
    byte_printer = Console(file=StringIO())
    byte_printer.out(bytes_array, end='')
    bytestr = byte_printer.file.getvalue()

    if bytestr.startswith("b'"):
        bytestr = bytestr.removeprefix("b'").removesuffix("'")
    elif bytestr.startswith('b"'):
        bytestr = bytestr.removeprefix('b"').removesuffix('"')
    else:
        raise RuntimeError(f"Unexpected byte string {bytestr}")

    return bytestr


def rich_text_view_of_raw_bytes(_bytes: bytes, bytes_match: BytesMatch) -> Text:
    """Print raw bytes to a Text object, highlighing the bytes in the bytes_match BytesMatch"""
    surrounding_bytes_str = clean_byte_string(_bytes)
    highlighted_bytes_str = clean_byte_string(bytes_match.bytes)
    highlighted_bytes_str_length = len(highlighted_bytes_str)
    highlight_idx = _find_str_rep_of_bytes(surrounding_bytes_str, highlighted_bytes_str, bytes_match)

    txt = Text(surrounding_bytes_str[:highlight_idx], style=GREY)
    matched_bytes_str = surrounding_bytes_str[highlight_idx:highlight_idx + highlighted_bytes_str_length]
    txt.append(matched_bytes_str, style=bytes_match.highlight_style)
    txt.append(surrounding_bytes_str[highlight_idx + highlighted_bytes_str_length:], style=GREY)
    return txt


def hex_view_of_raw_bytes(_bytes: bytes, bytes_match: BytesMatch) -> Text:
    hex_str = hex_text(_bytes)
    highlight_start_idx = bytes_match.highlight_start_idx * 3
    highlight_end_idx = bytes_match.highlight_end_idx * 3
    hex_str.stylize(bytes_match.highlight_style, highlight_start_idx, highlight_end_idx)
    lines = hex_str.wrap(console, HEX_CHARS_PER_LINE * 3)
    return newline_join([Text('  ').join(line.wrap(console, HEX_CHARS_PER_GROUP * 3)) for line in lines])


def ascii_view_of_raw_bytes(_bytes: bytes, bytes_match: BytesMatch) -> Text:
    txt = Text('', style=BYTES)

    for i, b in enumerate(_bytes):
        if i < bytes_match.highlight_start_idx or i > bytes_match.highlight_end_idx:
            style1 = 'color(246)'
            style2 = 'color(234)'
        else:
            style1 = None
            style2 = None

        _byte = b.to_bytes(1, byteorder)

        if b < 32:
            txt.append('*', style=style2 or BYTES_BRIGHTER)
        elif b < 127:
            txt.append(_byte.decode('UTF-8'), style1 or BYTES_BRIGHTEST)
        elif b <= 160:
            txt.append('*', style=style2 or BYTES_HIGHLIGHT)
        else:
            txt.append('*', style=style2 or BYTES)

    segments = [txt[i:i + HEX_CHARS_PER_GROUP] for i in range(0, len(txt), HEX_CHARS_PER_GROUP)]

    lines = [
        Text('  ').join(segments[i:min(len(segments), i + HEX_GROUPS_PER_LINE)])
        for i in range(0, len(segments), HEX_GROUPS_PER_LINE)
    ]

    return newline_join(lines)


def hex_text(_bytes: bytes) -> Text:
    return Text(hex_string(_bytes), style=GREY)


def hex_string(_bytes: bytes) -> str:
    return ' '.join([hex(b).removeprefix('0x').rjust(2, '0') for i, b in enumerate(_bytes)])


def print_bytes(bytes_array: bytes, style=None) -> None:
    """Convert bytes to a string representation and print to console"""
    for line in bytes_array.split(NEWLINE_BYTE):
        console.print(escape(clean_byte_string(line)), style=style or 'bytes')


def truncate_for_encoding(_bytes: bytes, encoding: str) -> bytes:
    """
    Truncate bytes to the a modulus of the char width of the given encoding.
    For utf-16 this means truncate to a multiple of 2, for utf-32 to a multiple of 4.
    """
    char_width = encoding_width(encoding)
    num_bytes = len(_bytes)
    num_extra_bytes = num_bytes % char_width

    if char_width <= 1 or num_bytes <= char_width or num_extra_bytes == 0:
        return _bytes
    else:
        return _bytes[:-num_extra_bytes]


def _find_str_rep_of_bytes(surrounding_bytes_str: str, highlighted_bytes_str: str, highlighted_bytes: BytesMatch):
    """
    Find the position of bytes_str in surrounding_byte_str. Both args are raw text dumps of binary data.
    Because strings are longer than bytes (stuff like '\xcc' are 4 chars when printed are one byte and the ANSI unprintables
    include stuff like 'NegativeAcknowledgement' which is over 20 chars) they represent so we have to re-find the location to highlight the bytes
    correctly.
    """
    # Start a few chars in to avoid errors: sometimes we're searching for 1 or 2 bytes and there's a false positive
    # in the extra bytes. Tthis isn't perfect - it's starting us at the first index into the *bytes* that's safe to
    # check but this is almost certainly far too soon given the large % of bytes that take 4 chars to print ('\x02' etc)
    highlight_idx = surrounding_bytes_str.find(highlighted_bytes_str, highlighted_bytes.highlight_start_idx)

    # TODO: Somehow \' and ' don't always come out the same :(
    if highlight_idx == -1:
        log.info(f"Failed to find highlighted_bytes in first pass so deleting single quotes and retrying. " + \
                  "Highlighting may be off by a few chars,")

        surrounding_bytes_str = surrounding_bytes_str.replace("\\'", "'")
        highlight_idx = surrounding_bytes_str.find(highlighted_bytes_str)

        if highlight_idx == -1:
            log.warning(f"Failed to find\n{highlighted_bytes_str}\nin surrounding bytes:\n{surrounding_bytes_str}")
            log.warning("Highlighting will not work on this decoded string.")

    return highlight_idx



"""
Methods for building Rich layout elements for display of results.
"""
import hashlib
from collections import namedtuple
from typing import Optional, Union

from rich.table import Column, Table

from yaralyzer.helpers.rich_text_helper import LEFT, size_text
from yaralyzer.output.rich_console import GREY

BytesInfo = namedtuple('BytesInfo', ['size', 'md5', 'sha1', 'sha256'])


def bytes_hashes_table(
        bytes_or_bytes_info: Union[bytes, BytesInfo],
        title: Optional[str] = None,
        title_justify: str = LEFT
    ) -> Table:
    """Build a table to show the MD5, SHA1, SHA256, etc."""
    if isinstance(bytes_or_bytes_info, bytes):
        bytes_info = compute_file_hashes(bytes_or_bytes_info)
    else:
        bytes_info = bytes_or_bytes_info

    table = Table(
        'Size',
        Column(size_text(bytes_info.size)),
        title=f" {title} Bytes Info" if title else None,
        title_style=GREY,
        title_justify=title_justify
    )
    table.add_row('MD5', bytes_info.md5)
    table.add_row('SHA1', bytes_info.sha1)
    table.add_row('SHA256', bytes_info.sha256)
    table.columns[1].style = 'orange3'
    table.columns[1].header_style = 'bright_cyan'
    return table


def compute_file_hashes(_bytes: bytes) -> BytesInfo:
    return BytesInfo(
        size=len(_bytes),
        md5=hashlib.md5(_bytes).hexdigest().upper(),
        sha1=hashlib.sha1(_bytes).hexdigest().upper(),
        sha256=hashlib.sha256(_bytes).hexdigest().upper()
    )


def compute_file_hashes_for_file(file_path) -> BytesInfo:
    with open(file_path, 'rb') as file:
        return compute_file_hashes(file.read())



"""
Class to  measure what we enounter as we iterate over every single match of a relatively simple byte level regex
(e.g. "bytes between quotes") against a relatively large pool of close to random encrypted binary data

Things like how much many of our matched bytes were we able to decode easily vs. by force vs. not at all,
were some encodings have a higher pct of success than others (indicating part of our mystery data might be encoded
that way?

TODO: use @dataclass decorator https://realpython.com/python-data-classes/
"""
from collections import defaultdict

from yaralyzer.decoding.bytes_decoder import BytesDecoder
from yaralyzer.util.logging import log


class RegexMatchMetrics:
    def __init__(self) -> None:
        self.match_count = 0
        self.bytes_matched = 0
        self.matches_decoded = 0
        self.easy_decode_count = 0
        self.forced_decode_count = 0
        self.undecodable_count = 0
        self.skipped_matches_lengths = defaultdict(lambda: 0)
        self.bytes_match_objs = []  # Keep a copy of all matches in memory
        self.per_encoding_stats = defaultdict(lambda: RegexMatchMetrics())

    def num_matches_skipped_for_being_empty(self) -> int:
        return self.skipped_matches_lengths[0]

    def num_matches_skipped_for_being_too_big(self) -> int:
        return sum({k: v for k, v in self.skipped_matches_lengths.items() if k > 0}.values())

    def tally_match(self, decoder: BytesDecoder) -> None:
        log.debug(f"Tallying {decoder.bytes_match} ({len(decoder.decodings)} decodings)")
        self.match_count += 1
        self.bytes_matched += decoder.bytes_match.match_length
        self.bytes_match_objs.append(decoder.bytes_match)

        if not decoder.bytes_match.is_decodable():
            self.skipped_matches_lengths[decoder.bytes_match.match_length] += 1

        for decoding_attempt in decoder.decodings:
            log.debug(f"Tallying decoding for {decoding_attempt.encoding}")
            encoding_stats = self.per_encoding_stats[decoding_attempt.encoding]

            if decoding_attempt.failed_to_decode:
                encoding_stats.undecodable_count += 1
            else:
                encoding_stats.match_count += 1
                encoding_stats.matches_decoded += 1

                if decoding_attempt.was_force_decoded:
                    encoding_stats.forced_decode_count += 1
                else:
                    encoding_stats.easy_decode_count += 1

    def __eq__(self, other):
        for k, v in vars(self).items():
            if v != vars(other)[k]:
                return False

        return True

    def __str__(self):
        return f"<matches: {self.match_count}, " + \
               f"bytes: {self.bytes_matched}, " + \
               f"decoded: {self.matches_decoded} " + \
               f"too_big: {self.num_matches_skipped_for_being_too_big()}, " + \
               f"empty: {self.num_matches_skipped_for_being_empty()}>" + \
               f"empty: {self.undecodable_count}>"



"""
Helper methods to work with strings.
"""
from functools import partial
from typing import Any, Callable, List

INDENT_DEPTH = 4
INDENT_SPACES = INDENT_DEPTH * ' '


def escape_yara_pattern(pattern: str) -> str:
    return pattern.replace('/', '\\/')


def line_count(_string: str) -> int:
    return len(_string.split("\n"))


def hex_to_string(_string: str) -> str:
    """String '0D 0A 25 25 45 4F 46 0D 0A' becomes '\r\n%%EOF\r\n'"""
    return bytearray.fromhex(_string.replace(' ', '')).decode()


def str_join(_list: List[Any], separator: str, func: Callable = str) -> str:
    """
    Return a comma separated list of strings. If func is provided the output of calling
    it on each element of the list will be used instead of str()
    """
    func = func or str
    return separator.join([func(item) for item in _list])


comma_join = partial(str_join, separator=', ')
newline_join = partial(str_join, separator='\n')
